# -*- coding: utf-8 -*-

from django.apps import apps
from django.core.management import call_command
from django.db.models.signals import (
    post_delete, post_migrate, post_save,
)
from django.dispatch import receiver

from navsy import cache
from navsy.models import Page, Route
from navsy.utils import i18n_utils


app_config = apps.get_app_config('navsy')


@receiver(post_migrate, sender=app_config, dispatch_uid='post_migrate_app')
def post_migrate_app(sender, **kwargs):
    cache.delete_routes()
    Page.create_default_page()

    if i18n_utils.use_modeltranslation():
        call_command('update_translation_fields')


@receiver(post_save, sender=Page, dispatch_uid='post_save_page')
def post_save_page(sender, instance, **kwargs):
    instance.create_default_route()
    instance.update_unique_home()


@receiver(post_save, sender=Route, dispatch_uid='post_save_route')
def post_save_route(sender, instance, **kwargs):
    cache.delete_routes()


@receiver(post_delete, sender=Route, dispatch_uid='post_delete_route')
def post_delete_route(sender, instance, **kwargs):
    cache.delete_routes()



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('navsy', '0006_removed_slug_uniqueness'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='page',
            options={'ordering': ['tn_order'], 'verbose_name': 'Page', 'verbose_name_plural': 'Pages'},
        ),
        migrations.RenameField(
            model_name='page',
            old_name='children_count',
            new_name='tn_children_count'
        ),
        migrations.RenameField(
            model_name='page',
            old_name='children_pks',
            new_name='tn_children_pks'
        ),
        migrations.RenameField(
            model_name='page',
            old_name='parent',
            new_name='tn_parent'
        ),
        migrations.RenameField(
            model_name='page',
            old_name='parents_count',
            new_name='tn_ancestors_count'
        ),
        migrations.RenameField(
            model_name='page',
            old_name='parents_pks',
            new_name='tn_ancestors_pks'
        ),
        migrations.RemoveField(
            model_name='page',
            name='priorities',
        ),
        migrations.RenameField(
            model_name='page',
            old_name='priority',
            new_name='tn_priority'
        ),
        migrations.RenameField(
            model_name='page',
            old_name='siblings_count',
            new_name='tn_siblings_count'
        ),
        migrations.RenameField(
            model_name='page',
            old_name='siblings_pks',
            new_name='tn_siblings_pks'
        ),
        migrations.AddField(
            model_name='page',
            name='tn_depth',
            field=models.PositiveSmallIntegerField(default=0, editable=False, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(10)], verbose_name='Depth'),
        ),
        migrations.AddField(
            model_name='page',
            name='tn_descendants_count',
            field=models.PositiveSmallIntegerField(default=0, editable=False, verbose_name='Descendants count'),
        ),
        migrations.AddField(
            model_name='page',
            name='tn_descendants_pks',
            field=models.CharField(blank=True, default='', editable=False, max_length=500, verbose_name='Descendants pks'),
        ),
        migrations.AddField(
            model_name='page',
            name='tn_index',
            field=models.PositiveSmallIntegerField(default=0, editable=False, verbose_name='Index'),
        ),
        migrations.AddField(
            model_name='page',
            name='tn_level',
            field=models.PositiveSmallIntegerField(default=0, editable=False, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(10)], verbose_name='Level'),
        ),
        migrations.AddField(
            model_name='page',
            name='tn_order',
            field=models.PositiveSmallIntegerField(default=0, editable=False, verbose_name='Order'),
        ),
        migrations.AlterField(
            model_name='page',
            name='tn_ancestors_count',
            field=models.PositiveSmallIntegerField(default=0, editable=False, verbose_name='Ancestors count'),
        ),
        migrations.AlterField(
            model_name='page',
            name='tn_ancestors_pks',
            field=models.CharField(blank=True, default='', editable=False, max_length=500, verbose_name='Ancestors pks'),
        ),
        migrations.AlterField(
            model_name='page',
            name='tn_children_count',
            field=models.PositiveSmallIntegerField(default=0, editable=False, verbose_name='Children count'),
        ),
        migrations.AlterField(
            model_name='page',
            name='tn_children_pks',
            field=models.CharField(blank=True, default='', editable=False, max_length=500, verbose_name='Children pks'),
        ),
        migrations.AlterField(
            model_name='page',
            name='tn_parent',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='tn_children', to='navsy.Page', verbose_name='Parent'),
        ),
        migrations.AlterField(
            model_name='page',
            name='tn_priority',
            field=models.PositiveSmallIntegerField(default=0, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(9999)], verbose_name='Priority'),
        ),
        migrations.AlterField(
            model_name='page',
            name='tn_siblings_count',
            field=models.PositiveSmallIntegerField(default=0, editable=False, verbose_name='Siblings count'),
        ),
        migrations.AlterField(
            model_name='page',
            name='tn_siblings_pks',
            field=models.CharField(blank=True, default='', editable=False, max_length=500, verbose_name='Siblings pks'),
        ),
    ]



# -*- coding: utf-8 -*-

from navsy.utils import path_utils

import re


__default_regexs = [

    # <blank>
    "",

    # <date_day>
    "(?P<date_day>((0|1|2)?([1-9])|[1-3]0|31))",

    # <date_month>
    "(?P<date_month>(0?([1-9])|10|11|12))",

    # <date_year>
    "(?P<date_year>[\d]{4})",

    # <date_year><date_month>
    "(?P<date_year>[\d]{4})(?P<date_month>(0?([1-9])|10|11|12))",

    # <date_year><date_month><date_day>
    "(?P<date_year>[\d]{4})(?P<date_month>(0?([1-9])|10|11|12))(?P<date_day>((0|1|2)?([1-9])|[1-3]0|31))",

    # <date_year>-<date_month>-<date_day>
    "(?P<date_year>[\d]{4})-(?P<date_month>(0?([1-9])|10|11|12))-(?P<date_day>((0|1|2)?([1-9])|[1-3]0|31))",

    # <date_year>/<date_month>
    "(?P<date_year>[\d]{4})/(?P<date_month>(0?([1-9])|10|11|12))",

    # <date_year>/<date_month>/<date_day>
    "(?P<date_year>[\d]{4})/(?P<date_month>(0?([1-9])|10|11|12))/(?P<date_day>((0|1|2)?([1-9])|[1-3]0|31))",

    # <email>
    "(?P<email>(([a-z\d]+([\.\_]{1}[a-z\d]+)?){6})\@([a-z\d]+([\.\_\-]{1}[a-z\d]+){1,}[a-z]+))",

    # <email_username>@<email_domain>
    "(?P<email_username>([a-z\d]+([\.\_]{1}[a-z\d]+)?){6})\@(?P<email_domain>[a-z\d]+([\.\_\-]{1}[a-z\d]+){1,}[a-z]+)",

    # <file_name>.<file_ext>
    "(?P<file_name>[\w-]+).(?P<file_ext>[\w]+)",

    # <id>
    "(?P<id>[\d]+)",

    # <id>/<slug>
    "(?P<id>[\d]+)/(?P<slug>[\w-]+)",

    # <page>
    "(?P<page>[\d]+)",

    # <slug>
    "(?P<slug>[\w-]+)",

    # <slug_parent>/<slug>
    "(?P<slug_parent>[\w-]+)/(?P<slug>[\w-]+)",

    # <username>
    "(?P<username>[\w.@-]+)",

    # <uuid>
    "(?P<uuid>[a-fA-F0-9]{8}-?[a-fA-F0-9]{4}-?[1345][a-fA-F0-9]{3}-?[a-fA-F0-9]{4}-?[a-fA-F0-9]{12})",

    # <version>
    "(?P<version>[\d]+(\.[\d]+)+)",

    # <version_major>.<version_minor>
    "(?P<version_major>[\d]+)\.(?P<version_minor>[\d]+)",

    # <version_major>.<version_minor>.<version_build>
    "(?P<version_major>[\d]+)\.(?P<version_minor>[\d]+)\.(?P<version_build>[\d]+)",

    # by-<slug>
    "by-(?P<slug>[\w-]+)",

    # login
    "login",

    # logout
    "logout",
]


def get_default_regexs():
    return list(__default_regexs)


def get_regexs():
    return get_default_regexs()


REGEX_DISPLAY_RE = re.compile(
    r'^([\w\-\_]+)((\.[\w]+)?)+|(\<\w+\>)|([\w\-\_\.\@\/]*)\(\?P|([\/\w\-\_\.]+)$')

# ([\w\-\_]+)               prefix
# ((\.[\w]+)?)+             extensions
# (\<\w+\>)                 parameter
# ([\w\-\_\.\@\/]*)\(\?P    parameters conjunction
# ([\/\w\-\_\.]+)           suffix


def get_regex_choices():

    return [[r, get_regex_display(r)] for r in get_regexs()]


def get_regex_display(regex):

    if not regex:
        return '<blank>'

    matches = REGEX_DISPLAY_RE.findall(regex)
    # print(matches)

    s = ''
    for m in matches:
        for i in m:
            if i:
                s += str(i)
    return s


def get_regex_format(regex):

    if not regex:
        return ''

    s = get_regex_display(regex)
    s = s.replace('<', '{')
    s = s.replace('>', '}')
    return s


def get_regex_re(base_url, regex):

    path = path_utils.fix_path(base_url, regex)
    if path.endswith('/'):
        pattern = r'^%s/?$' % path[0:-1]
    else:
        pattern = r'^%s$' % path
    obj = re.compile(pattern)
    return obj



# -*- coding: utf-8 -*-

from django.conf import settings
from django.contrib import admin
from django.db import models
from django.forms import TextInput, Textarea
from django.template.loader import get_template
from django.utils.translation import ugettext as _

from treenode.admin import TreeNodeModelAdmin

from navsy.forms import PageForm, RouteForm, RouteInlineForm
from navsy.models import Page, Route
from navsy.utils import import_function, i18n_utils, url_utils

__PageAdminBaseClass = TreeNodeModelAdmin

if i18n_utils.use_modeltranslation():

    from modeltranslation.admin import TabbedTranslationAdmin
    from modeltranslation.utils import get_translation_fields

    class __PageAdminBaseClass(TreeNodeModelAdmin, TabbedTranslationAdmin):
        pass


class RouteAdmin(admin.ModelAdmin):

    form = RouteForm

    list_display = ('url_display_with_redirect',
                    'view_name', 'view_function_path', 'view_check',
                    'response_type', 'is_enabled', 'priority', )
    list_editable = ('view_name', 'view_function_path', 'is_enabled', 'priority', )
    save_on_top = True

    def url_display_with_redirect(self, obj):

        obj_style = '' if obj.has_response() else 'opacity: 0.5;'
        obj_html = ''
        obj_html += '<span style="%s">' % obj_style
        obj_html += obj.get_absolute_url_format()

        redirect_url = obj.get_redirect_url()
        if redirect_url:
            redirect_label_style = 'color: #AAAAAA; font-weight: normal;'
            redirect_url_style = 'color: #888888; font-weight: normal;'
            redirect_html = '&searr; \
                <small style="%s">redirect: </small>\
                <small style="%s">%s</small>\
                ' % (redirect_label_style, redirect_url_style, redirect_url, )
            obj_html += '<br>%s' % (redirect_html, )

        obj_html += '</span>'
        return obj_html

    url_display_with_redirect.short_description = 'URL'
    url_display_with_redirect.allow_tags = True

    def response_type(self, obj):

        badge_style = u'background-color: #CCCCCC; \
                        color: #FFFFFF; \
                        text-transform: uppercase; \
                        padding: 2px 6px; \
                        border-radius: 4px; \
                        margin-right: 5px;'

        badge_format = u'<small style="%s">%s</small>'

        if obj.view_function_path:
            return badge_format % (badge_style, _('function'), )

        elif obj.view_template_path:
            return badge_format % (badge_style, _('template'), )

        elif obj.get_redirect_url():
            return badge_format % (badge_style, _('redirect'), )
        else:
            return None

    response_type.allow_tags = True
    response_type.short_description = 'Response Type'

    def view_check(self, obj):

        if obj.view_function_path:
            return import_function(obj.view_function_path) is not None

        elif obj.view_template_path:
            try:
                view_template = get_template(obj.view_template_path)
                return True
            except Exception:
                return False

        return None

    view_check.short_description = 'View Check'
    view_check.boolean = True

    def get_fieldsets(self, request, obj=None):

        fieldsets = (
            (None, {
                'classes': ('wide', ),
                'fields': (
                    'page', 'regex', 'is_enabled',
                )
            }),
            ('Redirect', {
                'classes': ('wide', ),
                'fields': (
                    'redirect_to_page', 'redirect_to_path', 'redirect_to_url',
                )
            }),
            ('View', {
                'classes': ('wide', ),
                'fields': (
                    'view_name', 'view_function_path', 'view_template_path',
                )
            }),
        )

        return fieldsets

    def get_changelist_form(self, request, **kwargs):
        return RouteForm

admin.site.register(Route, RouteAdmin)


class RouteInline(admin.TabularInline):

    model = Route
    form = RouteInlineForm
    fk_name = 'page'
    extra = 0
    fields = ('regex', 'view_function_path', 'view_name',
              'is_enabled', 'priority', 'route_admin_link', )
    readonly_fields = ('route_admin_link', )
    view_on_site = False
    show_change_link = True

    def route_admin_link(self, obj):
        link_style = 'padding:6px 12px; line-height:25px; '
        if obj.id:
            link_href = url_utils.get_admin_change_url(obj)
        else:
            link_href = '#'
            link_style += 'opacity:0.2; pointer-events:none;'
        return u'<a href="%s" target="_self" class="button" style="%s">%s</a>\
                ' % (link_href, link_style, _('Change'), )

    route_admin_link.allow_tags = True
    route_admin_link.short_description = 'Options'


class PageAdmin(__PageAdminBaseClass):

    treenode_accordion = True

    form = PageForm

    def get_changelist_form(self, request, **kwargs):
        return PageForm

    def status_check(self, obj):
        if obj.is_draft:
            return None
        elif obj.is_published:
            return True
        elif obj.is_unpublished:
            return False
        else:
            return False

    status_check.short_description = 'Status Check'
    status_check.boolean = True

    list_display = ('url_with_link', 'status', 'status_check', 'is_home', 'in_menu', 'tn_priority', )
    list_editable = ('is_home', 'in_menu', 'status', 'tn_priority', )
    list_per_page = 100
    save_on_top = True

    def url_with_link(self, obj):
        obj_route = obj.get_route()
        obj_style = '' if obj_route and obj_route.has_response() else 'opacity: 0.5;'
        obj_url = obj.get_absolute_url()
        return u'<a style="%s" href="%s">%s</a>' % (obj_style, obj_url, obj_url, )

    url_with_link.short_description = 'URL'
    url_with_link.allow_tags = True

    inlines = [RouteInline]

    formfield_overrides = {
        models.TextField: {'widget': Textarea(attrs={'rows':2, 'cols':80})},
    }

    def get_fieldsets(self, request, obj=None):

        if i18n_utils.use_modeltranslation():
            none_fields = ('tn_parent', ) + tuple(get_translation_fields('name'))
            seo_fields = tuple(get_translation_fields('seo_title')) + \
                         tuple(get_translation_fields('seo_description')) + \
                         tuple(get_translation_fields('seo_keywords'))
        else:
            none_fields = ('tn_parent', 'name', 'status', )
            seo_fields = ('seo_title', 'seo_description', 'seo_keywords', )

        sitemap_fields = ('sitemap_item', 'sitemap_priority', 'sitemap_changefreq', )

        fieldsets = (
            (None, {
                'classes': ('wide', ),
                'fields': none_fields
            }),
            ('SEO'.upper(), {
                'classes': ('collapse', ),
                'fields': seo_fields
            }),
            ('Sitemap'.upper(), {
                'classes': ('collapse', ),
                'fields': sitemap_fields
            }),
        )

        return fieldsets

    def save_related(self, request, form, formsets, change):
        super(PageAdmin, self).save_related(request, form, formsets, change)
        page_obj = form.instance
        page_obj.create_default_route()

admin.site.register(Page, PageAdmin)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('navsy', '0015_fixed_route_default_ordering'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='page',
            name='published',
        ),
    ]



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('navsy', '0013_remove_route_redirect_to_first_child_page'),
    ]

    operations = [
        migrations.AddField(
            model_name='page',
            name='status',
            field=models.PositiveSmallIntegerField(choices=[(1, 'Draft'), (2, 'Published'), (0, 'Unpublished')], default=1),
        ),
    ]



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('navsy', '0010_added_regex_fields_to_route'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='route',
            unique_together=set([('page', 'regex')]),
        ),
        migrations.RemoveField(
            model_name='route',
            name='pattern',
        ),
        migrations.DeleteModel(
            name='Pattern',
        ),
    ]



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.conf import settings
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import render
from django.utils.encoding import python_2_unicode_compatible, force_text
try:
    from slugify import slugify
except ImportError:
    from django.utils.text import slugify

from treenode.cache import query_cache
from treenode.models import TreeNodeModel

from navsy.utils import import_function, path_utils, pattern_utils, url_utils


class Page(TreeNodeModel):

    treenode_display_field = 'name'

    def __init__(self, *args, **kwargs):
        super(Page, self).__init__(*args, **kwargs)

    STATUS_DRAFT = 1
    STATUS_PUBLISHED = 2
    STATUS_UNPUBLISHED = 0
    STATUS_CHOICES = (
        (STATUS_DRAFT, 'Draft', ),
        (STATUS_PUBLISHED, 'Published', ),
        (STATUS_UNPUBLISHED, 'Unpublished', ),
    )
    SITEMAP_PRIORITY_CHOICES = (
        ('0.0', '0.0', ),
        ('0.1', '0.1', ),
        ('0.2', '0.2', ),
        ('0.3', '0.3', ),
        ('0.4', '0.4', ),
        ('0.5', '0.5', ),
        ('0.6', '0.6', ),
        ('0.7', '0.7', ),
        ('0.8', '0.8', ),
        ('0.9', '0.9', ),
        ('1.0', '1.0', ),
    )
    SITEMAP_CHANGEFREQ_CHOICES = (
        ('always', 'always', ),
        ('hourly', 'hourly', ),
        ('daily', 'daily', ),
        ('weekly', 'weekly', ),
        ('monthly', 'monthly', ),
        ('yearly', 'yearly', ),
        ('never', 'never', ),
    )

    SITEMAP_DEFAULT_PRIORITY = '0.5'
    SITEMAP_DEFAULT_CHANGEFREQ = 'never'

    name = models.CharField(max_length=100)
    slug = models.CharField(max_length=100, blank=True, default='')

    is_home = models.BooleanField(default=False)
    in_menu = models.BooleanField(default=True)

    status = models.PositiveSmallIntegerField(
        choices=STATUS_CHOICES, default=STATUS_DRAFT)

    seo_title = models.CharField(blank=True, max_length=100)
    seo_description = models.TextField(blank=True)
    seo_keywords = models.TextField(blank=True)

    sitemap_item = models.BooleanField(default=True)
    sitemap_priority = models.CharField(
        max_length=5,
        choices=SITEMAP_PRIORITY_CHOICES,
        default=SITEMAP_DEFAULT_PRIORITY,
        blank=True)
    sitemap_changefreq = models.CharField(
        max_length=50,
        choices=SITEMAP_CHANGEFREQ_CHOICES,
        default=SITEMAP_DEFAULT_CHANGEFREQ,
        blank=True)

    @staticmethod
    def create_default_page():
        if Page.objects.count() == 0:
            page_obj = Page.objects.create(name='Home', is_home=True, in_menu=False)
            page_obj.create_default_route()

    def create_default_route(self):
        return Route.objects.get_or_create(page=self, regex='')

    def update_unique_home(self):
        if self.is_home:
            Page.objects.exclude(pk=self.pk).update(is_home=False)

    def get_absolute_url(self):
        return url_utils.reverse_url(self.get_path())

    def get_path(self):
        if self.is_home:
            path = ''
        else:
            breadcrumbs = self.get_breadcrumbs()
            slugs = [page.slug for page in breadcrumbs if not page.is_home]
            path = path_utils.join_path_slugs(slugs)
        path = path_utils.fix_path(path)
        return path

    def get_route(self):
        from navsy import cache
        return cache.get_route_by_page(self)

    @property
    def is_draft(self):
        statuses = self.get_breadcrumbs('status')
        statuses_flags = [(status == Page.STATUS_DRAFT) for status in statuses]
        return any(statuses_flags) and not self.is_unpublished

    @property
    def is_published(self):
        statuses = self.get_breadcrumbs('status')
        statuses_flags = [(status == Page.STATUS_PUBLISHED) for status in statuses]
        return all(statuses_flags)

    @property
    def is_unpublished(self):
        statuses = self.get_breadcrumbs('status')
        statuses_flags = [(status == Page.STATUS_UNPUBLISHED) for status in statuses]
        return any(statuses_flags)

    @property
    def path(self):
        return self.get_path()

    @property
    def route(self):
        return self.get_route()

    @property
    def url(self):
        return self.get_absolute_url()

    def save(self, *args, **kwargs):

        page_obj = self
        page_obj.slug = slugify(page_obj.name)

        for lang_code, lang_display in settings.LANGUAGES:
            lang_code = lang_code.replace('-', '_').lower()

            name_field = 'name_%s' % (lang_code, )
            slug_field = 'slug_%s' % (lang_code, )

            if hasattr(page_obj, name_field) and hasattr(page_obj, slug_field):
                name_value = str(getattr(page_obj, name_field, ''))
                slug_value = slugify(name_value)
                setattr(page_obj, slug_field, slug_value)

        super(Page, page_obj).save(*args, **kwargs)

    class Meta(TreeNodeModel.Meta):
        verbose_name = 'Page'
        verbose_name_plural = 'Pages'


@python_2_unicode_compatible
class Route(models.Model):

    page = models.ForeignKey(
        'navsy.Page', blank=True, null=True,
        on_delete=models.CASCADE, related_name='routes')

    REGEX_CHOICES = pattern_utils.get_regex_choices()
    regex = models.CharField(
        max_length=255, blank=True, default='', choices=REGEX_CHOICES)

    view_name = models.CharField(
        max_length=100, unique=True, blank=True, null=True, db_index=True,
        help_text='(unique name used for reverse URL)')
    view_function_path = models.CharField(
        max_length=100, blank=True, help_text='(ex. app.views.view_func)')
    view_template_path = models.CharField(
        max_length=100, blank=True, help_text='(ex. app/template.html)')

    redirect_to_page = models.ForeignKey(
        'navsy.Page', blank=True, null=True, on_delete=models.SET_NULL)
    redirect_to_path = models.CharField(max_length=200, blank=True)
    redirect_to_url = models.URLField(blank=True)

    is_enabled = models.BooleanField(default=True)

    priority = models.PositiveSmallIntegerField(
        default=0, validators=[MinValueValidator(0), MaxValueValidator(9999)])

    def get_absolute_base_url(self):
        page_obj = self.get_page()
        return page_obj.get_path() if page_obj else ''

    def get_absolute_url_format(self):
        return path_utils.fix_path(
            self.get_absolute_base_url(), self.get_regex_format())

    def get_absolute_url_regex(self):
        return pattern_utils.get_regex_re(
            self.get_absolute_base_url(), self.regex)

    def get_absolute_url(self, parameters=None):
        return url_utils.reverse_url(
            url_utils.format_url(
                self.get_absolute_url_format(), parameters))

    def get_page(self):
        return query_cache(Page, pk=self.page_id) if self.page_id else None

    def get_redirect_url(self):
        redirect_url = None
        if self.redirect_to_page:
            redirect_url = self.redirect_to_page.get_absolute_url()
        elif self.redirect_to_url:
            redirect_url = self.redirect_to_url
        elif self.redirect_to_path:
            if self.redirect_to_path.startswith('/'):
                redirect_path = self.redirect_to_path
            else:
                page_obj = self.get_page()
                redirect_path = page_obj.get_absolute_url() if page_obj else '/'  # base_url
            redirect_url = path_utils.fix_path(
                redirect_path, self.redirect_to_path)
        # if redirect_url:
        #    print(redirect_url)
        return redirect_url

    def get_regex_display(self):
        return pattern_utils.get_regex_display(self.regex)

    def get_regex_format(self):
        return pattern_utils.get_regex_format(self.regex)

    def get_response(self, request, *args, **kwargs):

        if not self.has_response(request):
            raise Http404

        redirect_url = self.get_redirect_url()
        if redirect_url:
            return HttpResponseRedirect(redirect_url)

        data = {
            'navsy_page': self.get_page(),
            'navsy_route': self,
            'navsy_route_args': args,
            'navsy_route_kwargs': kwargs,
        }

        request.navsy_data = data

        if self.view_template_path:
            return render(request, self.view_template_path)

        if self.view_function_path:
            view_function = import_function(self.view_function_path)
            if view_function:
                if kwargs:
                    view_response = view_function(request, **kwargs)
                elif args:
                    view_response = view_function(request, *args)
                else:
                    view_response = view_function(request)

                if isinstance(view_response, HttpResponse):
                    return view_response
                else:
                    raise Exception('View function "%s" doesn\'t return \
                        an HttpResponse object.' % (self.view_function_path, ))
            else:
                # this should never happen since there is form validation
                raise Exception('View function "%s" is not \
                    a valid function.' % (self.view_function_path, ))

        if settings.DEBUG:
            return render(request, 'navsy/default.html')
        else:
            return HttpResponse('')

    def has_response(self, request=None):
        if self.is_enabled:
            page_obj = self.get_page()
            if page_obj:
                if request:
                    user = getattr(request, 'user', None)
                    if user:
                        if user.is_staff or user.is_superuser:
                            return page_obj.is_published or page_obj.is_draft
                return page_obj.is_published
            return True
        else:
            return False

    @property
    def redirect_url(self):
        return self.get_redirect_url()

    @property
    def regex_display(self):
        return self.get_regex_display()

    @property
    def regex_format(self):
        return self.get_regex_format()

    @property
    def url(self):
        return self.get_absolute_url()

    def save(self, *args, **kwargs):
        if self.view_name:
            self.view_name = self.view_name.strip()
        if not self.view_name:
            self.view_name = None
        super(Route, self).save(*args, **kwargs)

    class Meta:
        unique_together = (('page', 'regex', ), )
        ordering = ['page', '-priority', 'regex']
        # ordering = ['-priority', 'regex', 'page']
        verbose_name = 'Route'
        verbose_name_plural = 'Routes'

    def __str__(self):
        return force_text(self.get_absolute_url_format())



# -*- coding: utf-8 -*-

from django.http import Http404, HttpResponseRedirect

from navsy.cache import get_routes
# from navsy.models import Route
from navsy.utils import path_utils, url_utils


def router(request, path=None):

    path, path_redirect = path_utils.fix_path_plus(path)
    if path_redirect:
        return HttpResponseRedirect(url_utils.reverse_url(path))
    # print(path)

    # routes_list = list(Route.objects.all())
    routes_list = get_routes()
    routes_list_sorted = [route_obj for route_obj in routes_list if route_obj.regex == ''] + \
                         [route_obj for route_obj in routes_list if route_obj.regex != '']
    routes_list = routes_list_sorted
    # for route_obj in routes_list:
    #     print(route_obj)

    route_not_found = Http404

    for route_obj in routes_list:
        route_re = route_obj.get_absolute_url_regex()
        route_match = route_re.match(path)
        # print('route match: %s -> %s == %s' % (
        #     True if route_match else False, path, route_re.regex_display, ))

        if route_match:
            route_match_kwargs = route_match.groupdict()
            route_match_keys = list(route_match_kwargs.keys())
            route_match_keys.sort(key=lambda arg: route_re.pattern.find(arg))
            route_match_args = [route_match_kwargs[key]
                                for key in route_match_keys]
            # print(route_match_kwargs)
            # print(route_match_keys)
            # print(route_match_args)

            try:
                route_response = route_obj.get_response(
                    request, *route_match_args, **route_match_kwargs)
                return route_response

            except Http404 as route_not_found:
                continue

    raise route_not_found



# -*- coding: utf-8 -*-

from django.core.cache import cache

from navsy.models import Route


def clear():
    delete_routes()


def delete_routes():
    cache.delete('navsy_routes')


def get_routes():
    routes_list = cache.get('navsy_routes', None)
    if not routes_list:
        routes_list = list(Route.objects.all())
        cache.set('navsy_routes', routes_list)
    return routes_list


def get_route_by_page(page):
    if not page or not page.pk:
        return None
    routes_list = get_routes()
    for route_obj in routes_list:
        if route_obj.regex == '':
            page_obj = route_obj.get_page()
            if page_obj.pk == page.pk:
                return route_obj
    return None


def get_route_by_view_name(view_name):
    if not view_name:
        return None
    routes_list = get_routes()
    for route_obj in routes_list:
        if route_obj.view_name == view_name:
            return route_obj
    return None



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('navsy', '0016_remove_page_published'),
    ]

    operations = [
        migrations.RenameField(
            model_name='route',
            old_name='published',
            new_name='is_enabled',
        ),
        migrations.RenameField(
            model_name='page',
            old_name='home',
            new_name='is_home',
        ),
        migrations.RenameField(
            model_name='page',
            old_name='menu',
            new_name='in_menu',
        ),
    ]



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import autoslug.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('navsy', '0004_page_seo'),
    ]

    operations = [
        migrations.AlterField(
            model_name='page',
            name='slug',
            field=autoslug.fields.AutoSlugField(blank=True, editable=False, null=True, populate_from='name', unique_with=('parent',)),
        ),
    ]



# -*- coding: utf-8 -*-

from django.apps import AppConfig
from django.db.utils import OperationalError


class NavsyConfig(AppConfig):

    name = 'navsy'
    verbose_name = 'Navigation'

    def ready(self):

        from navsy import cache, settings, signals
        from navsy.models import Page

        settings.check_settings()
        cache.clear()

        try:
            Page.update_tree()
        except OperationalError:
            pass



# -*- coding: utf-8 -*-

from navsy.models import Page
from navsy.utils import i18n_utils


if i18n_utils.use_modeltranslation():

    try:
        from modeltranslation.translator import translator, TranslationOptions

        class PageTranslationOptions(TranslationOptions):
            fields = (
                'name', 'slug',
                'seo_title', 'seo_description', 'seo_keywords', )

        translator.register(Page, PageTranslationOptions)

    except ImportError:
        pass


# -*- coding: utf-8 -*-

import django

if django.VERSION < (2, 0):
    from django.conf.urls import url as re_path
else:
    from django.urls import re_path

from navsy.cache import get_route_by_view_name
from navsy.exceptions import NoReverseMatch
# from navsy.models import Route
from navsy.views import router


def reverse(view_name, *args, **kvargs):

    if args and kvargs:
        raise Exception('reverse cannot receive both *args and **kvargs')
    elif args:
        parameters = list(args)
    elif kvargs:
        parameters = dict(kvargs)
    else:
        parameters = None

    # print('reverse parameters: %s' % (parameters, ))

    # try:
    #     route_obj = Route.objects.get(view_name=view_name)
    #     return route_obj.get_absolute_url(parameters=parameters)
    #
    # except Route.DoesNotExist:
    #     raise NoReverseMatch()

    route_obj = get_route_by_view_name(view_name)
    if route_obj:
        return route_obj.get_absolute_url(parameters=parameters)
    else:
        raise NoReverseMatch()

urlpatterns = [
    re_path(r'(?:(?P<path>.*))?$', router, name='navsy-router'),
]



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('navsy', '0008_removed_autoslugfield'),
    ]

    operations = [
        migrations.AlterField(
            model_name='page',
            name='tn_level',
            field=models.PositiveSmallIntegerField(default=1, editable=False, validators=[django.core.validators.MinValueValidator(1), django.core.validators.MaxValueValidator(10)], verbose_name='Level'),
        ),
    ]



# -*- coding: utf-8 -*-

import django

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

if django.VERSION < (1, 10):
    from django.core.urlresolvers import reverse
else:
    from django.urls import reverse

from navsy.exceptions import NoReverseMatch
from navsy.utils import i18n_utils


def check_settings():

    try:
        reverse('navsy-router')
    except NoReverseMatch:
        e = 'You must add "'
        e += 'url(r\'^\', include(\'navsy.urls\'))'
        e += '" to your urls.'
        raise ImproperlyConfigured(e)

    if i18n_utils.is_multilanguage():
        try:
            reverse('set_language')
        except NoReverseMatch:
            e = 'You must add "'
            e += 'url(r\'^i18n/\', include(\'django.conf.urls.i18n\'))'
            e += '" to your urls.'
            raise ImproperlyConfigured(e)



# -*- coding: utf-8 -*-

from django.conf import settings
from django.utils import translation

import re


def get_current_language():
    current_language = translation.get_language() or settings.LANGUAGE_CODE
    return current_language


def is_multilanguage():
    return bool(len(settings.LANGUAGES) > 1)


def split_languages(languages_codes_str):
    return re.findall(r'[\w\-]+', languages_codes_str.lower()
                      ) if languages_codes_str else []


def use_modeltranslation():
    if is_multilanguage():
        if 'modeltranslation' in settings.INSTALLED_APPS:
            try:
                import modeltranslation
                return True
            except ImportError:
                pass
    return False




# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('navsy', '0012_removed_regex_extra_fields'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='route',
            name='redirect_to_first_child_page',
        ),
    ]



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('navsy', '0014_page_status'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='route',
            options={'ordering': ['page', '-priority', 'regex'], 'verbose_name': 'Route', 'verbose_name_plural': 'Routes'},
        ),
    ]



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('navsy', '0007_added_treenode'),
    ]

    operations = [
        migrations.AlterField(
            model_name='page',
            name='slug',
            field=models.CharField(blank=True, default='', max_length=100),
        ),
    ]



# -*- coding: utf-8 -*-

from django.utils.module_loading import import_string


def import_function(path):
    try:
        func = import_string(path)
        if hasattr(func, '__call__'):
            return func
        else:
            return None
    except ImportError:
        return None
